// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the GNU General Public License version 2 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.core;

// FIXME: are traps on x/y meaningful?
// FIXME: trap on visible, trigger when parent visibility changes
// FIXME: mouse move/release still needs to propagate to boxen in which the mouse was pressed and is still held down
// FEATURE: reintroduce surface.abort

// Broken:
// - textures
// - align/origin
// - clipping (all forms)
// - mouse events
// - fonts/text
// - vertical layout

// - stroke clipping

import java.util.*;
import org.ibex.js.*;
import org.ibex.util.*;
import org.ibex.plat.*;
import org.ibex.graphics.*;

/**
 *  <p>
 *  Encapsulates the data for a single Ibex box as well as all layout
 *  rendering logic.
 *  </p>
 *
 *  <p>The rendering process consists of four phases; each requires
 *     one DFS pass over the tree</p>
 *  <ol><li> <b>pack()</b>: each box sets its childrens' row/col
 *  <ol><li> <b>constrain()</b>: contentwidth is computed
 *      <li> <b>resize()</b>: width/height and x/y positions are set
 *      <li> <b>render()</b>: children draw their content onto the PixelBuffer.
 *  </ol>
 *
 *  The first three passes together are called the <i>reflow</i>
 *  phase.  Reflowing is done in a seperate pass since SizeChanges
 *  trigger a Surface.abort; if rendering were done in the same pass,
 *  rendering work done prior to the Surface.abort would be wasted.
 */
public final class Box extends JS.Obj implements Callable, Mesh.Chain {

    public Mesh.Chain getMeshChainParent() { return parent; }
    public Mesh       getMesh()            { return mesh; }
    public Affine     getAffine()          { return transform; }

    // Macros //////////////////////////////////////////////////////////////////////

    final void REPLACE()     { for(Box b2 = this; b2 != null && !b2.test(REPLACE); b2 = b2.parent) b2.set(REPLACE); }
    final void RECONSTRAIN() { for(Box b2 = this; b2 != null && !b2.test(RECONSTRAIN); b2 = b2.parent) b2.set(RECONSTRAIN); }

    //#define CHECKSET_SHORT(prop) short nu = (short)JSU.toInt(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_INT(prop) int nu = JSU.toInt(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_FLAG(flag) boolean nu=JSU.toBoolean(value);if(nu == test(flag)) break; if (nu) set(flag); else clear(flag);
    //#define CHECKSET_BOOLEAN.N(prop) boolean nu = JSU.toBoolean(value); if (nu == prop) break; prop = nu;
    //#define CHECKSET_STRING(prop) if((value==null&&prop==null)||(value!=null&&JSU.toString(value).equals(prop)))break;prop=JSU.toString(value);

    // Instance Data //////////////////////////////////////////////////////////////////////

    private Box          parent       = null;
    private Box          redirect     = this;
    private int          flags        = VISIBLE | RECONSTRAIN | REPLACE | STOP_UPWARD_PROPAGATION | CLIP | MOVED;
    private BalancedTree bt           = null;
    private String       text         = null;
    private Font         font         = DEFAULT_FONT; 
    private Picture      texture      = null;
    public  int          fillcolor    = 0x00000000;
    private int          strokecolor  = 0xFF000000;
    public  float        flex         = 1;
    private Path         path         = null;
    private Path         clippath     = null;
    private Affine       transform    = Affine.identity();

    // FEATURE: polygon caching
    private Mesh      polygon      = null;
    private Mesh         mesh         = null;

    // specified directly by user
    public int minwidth = 0;
    public int maxwidth = Integer.MAX_VALUE;
    public int minheight = 0;
    public int maxheight = Integer.MAX_VALUE;

    // computed during reflow
    //public int width = 0;   // AS MEASURED IN PARENT SPACE!
    //public int height = 0;  // AS MEASURED IN PARENT SPACE!
    float _width;
    float _height;
    private int width;
    private int height;
    private int rootwidth;
    private int rootheight;
    public int getRootWidth() { return rootwidth; }
    public int getRootHeight() { return rootheight; }

    public int contentwidth = 0;      // == max(minwidth, textwidth, sum(child.contentwidth))
    public int contentheight = 0;


    // Instance Methods /////////////////////////////////////////////////////////////////////

    /** invoked when a resource needed to render ourselves finishes loading */
    public Object run(Object o) throws JSExn {
        if (texture == null) { Log.warn(Box.class, "perform() called with null texture"); return null; }
        if (texture.isLoaded) {
            setWidth(max(texture.width, minwidth), maxwidth); 
            setHeight(max(texture.height, minheight), maxheight); 
            dirty(); }
        else { JS res = texture.stream; texture = null; throw new JSExn("image not found: "+res.unclone()); }
        return null;
    }

    /** Adds the intersection of (x,y,w,h) and the node's current actual geometry to the Surface's dirty list */
    public void dirty() { if (path==null) dirty(0, 0, contentwidth, contentheight); else dirty(path); }
    public void dirty(int x, int y, int w, int h) { }
    public void dirty(Path p) {
        Affine a = transform.copy();
        for(Box cur = this; cur != null; cur = cur.parent) a.premultiply(cur.transform);
        long hbounds = p.horizontalBounds(a);
        long vbounds = p.verticalBounds(a);
        int x1 = (int)Encode.longToFloat2(hbounds);
        int x2 = (int)Encode.longToFloat1(hbounds);
        int y1 = (int)Encode.longToFloat2(vbounds);
        int y2 = (int)Encode.longToFloat1(vbounds);
        if (getSurface() != null) getSurface().dirty(x1, y1, x2-x1, y2-y1);
    }


    // Reflow ////////////////////////////////////////////////////////////////////////////////////////

    /** should only be invoked on the root box */
    public void reflow() {
        constrain(this, Affine.identity(), new BoundingBox());
        width  = rootwidth  = maxwidth;
        height = rootheight = maxheight;
        transform.e = 0;
        transform.f = 0;
        place(0, 0, width, height, false);
    }

    public float minWidth()      { return Encode.longToFloat1(transform.rotateBox(minwidth, minheight)); }
    public float minHeight()     { return Encode.longToFloat2(transform.rotateBox(minwidth, minheight)); }
    public float maxWidth()      { return Encode.longToFloat1(transform.rotateBox(maxwidth, maxheight)); }
    public float maxHeight()     { return Encode.longToFloat2(transform.rotateBox(maxwidth, maxheight)); }
    public float contentWidth()  { return Encode.longToFloat1(transform.rotateBox(contentwidth, contentheight)); }
    public float contentHeight() { return Encode.longToFloat2(transform.rotateBox(contentwidth, contentheight)); }

    /** used (single-threadedly) in constrain() */
    private static int xmin = 0, ymin = 0, xmax = 0, ymax = 0;

    private static class BoundingBox {
        public int xmin, ymin, xmax, ymax;
        public boolean unbounded() {
            return xmin==Integer.MAX_VALUE||xmax==Integer.MIN_VALUE||ymin==Integer.MAX_VALUE||ymax==Integer.MIN_VALUE; }
        public void reset() {
            xmin = Integer.MAX_VALUE; ymin = Integer.MAX_VALUE;
            xmax = Integer.MIN_VALUE; ymax = Integer.MIN_VALUE;
        }
        public void include(Affine a, float cw, float ch) {
            //#repeat contentwidth/contentheight contentheight/contentwidth minwidth/minheight row/col col/row \
            //        textwidth/textheight maxwidth/maxheight bounds/boundsy x1/y1 x2/y2 z1/q1 z2/q2 z3/q3 z4/q4 \
            //        horizontalBounds/verticalBounds e/f multiply_px/multiply_py xmin/ymin xmax/ymax
            float z1    = a.multiply_px(0, 0);
            float z2    = a.multiply_px(cw, ch);
            float z3    = a.multiply_px(cw, 0);
            float z4    = a.multiply_px(0, ch);
            xmin = min(xmin, (int)min(min(z1, z2), min(z3, z4)));
            xmax = max(xmax, (int)max(max(z1, z2), max(z3, z4)));
            //#end
        }
        public void include(Affine a, Path path) {
            //#repeat contentwidth/contentheight contentheight/contentwidth minwidth/minheight row/col col/row \
            //        textwidth/textheight maxwidth/maxheight bounds/boundsy x1/y1 x2/y2 z1/q1 z2/q2 z3/q3 z4/q4 \
            //        horizontalBounds/verticalBounds e/f multiply_px/multiply_py xmin/ymin xmax/ymax
            long bounds = path.horizontalBounds(a);
            float z1    = Encode.longToFloat2(bounds);
            float z2    = Encode.longToFloat1(bounds);
            float z3    = Encode.longToFloat2(bounds);
            float z4    = Encode.longToFloat1(bounds);
            xmin = min(xmin, (int)min(min(z1, z2), min(z3, z4)));
            xmax = max(xmax, (int)max(max(z1, z2), max(z3, z4)));
            //#end
        }
    }

    /** expand the {x,y}{min,max} boundingbox in space <tt>a</tt> to include this box */
    public void constrain(Box b, Affine a, BoundingBox bbox) {
        contentwidth = 0;
        contentheight = 0;
        a = a.copy().premultiply(transform);

        BoundingBox bbox2 = new BoundingBox();
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            child.constrain(this, Affine.identity(), bbox2);
            if (bbox2.unbounded()) { /* FIXME: why? */ bbox2.reset(); continue; }
            if (packed()) {
                // packed boxes mush together their childrens' bounding boxes
                if (test(XAXIS)) {
                    contentwidth  = contentwidth + (bbox2.xmax-bbox2.xmin);
                    contentheight = max(bbox2.ymax-bbox2.ymin, contentheight);
                } else {
                    contentwidth  = max(bbox2.xmax-bbox2.xmin, contentwidth);
                    contentheight = contentheight + (bbox2.ymax-bbox2.ymin);
                }
                bbox2.reset();
            }
        }
        if (!packed()) {
            // unpacked boxes simply use the "cumulative" bounding box
            contentwidth  = bbox2.xmax-bbox2.xmin;
            contentheight = bbox2.ymax-bbox2.ymin;
        }

        contentwidth = bound(minwidth, contentwidth, maxwidth);
        contentheight = bound(minheight, contentheight, maxheight);
        bbox.include(a, contentwidth, contentheight);
        if (path!=null) bbox.include(a, path);
    }

    void place(float x, float y, float w, float h, boolean keep) {
        int oldw = width;
        int oldh = height;
        width  = bound(contentwidth,  (int)Encode.longToFloat1(transform.inverse().rotateBox(w, h)), test(HSHRINK)?contentwidth:maxwidth);
        height = bound(contentheight, (int)Encode.longToFloat2(transform.inverse().rotateBox(w, h)), test(VSHRINK)?contentheight:maxheight);
        if (oldw!=width || oldh!=height) mesh = null;
        if (!keep) {
            Affine a = transform;
            transform.e = 0;
            transform.f = 0;
            float e;
            float f;
            //#repeat e/f x/y multiply_px/multiply_py horizontalBounds/verticalBounds bounds/boundsy z1/z1y z2/z2y z3/z3y z4/z4y
            long bounds = path==null ? 0                            : path.horizontalBounds(transform);
            float z1    = path==null ? a.multiply_px(0, 0)          : Encode.longToFloat2(bounds);
            float z2    = path==null ? a.multiply_px(width, height) : Encode.longToFloat1(bounds);
            float z3    = path==null ? a.multiply_px(width, 0)      : Encode.longToFloat2(bounds);
            float z4    = path==null ? a.multiply_px(0,     height) : Encode.longToFloat1(bounds);
            e = (-1 * min(min(z1, z2), min(z3, z4))) + x;
            //#end
            transform.e = e;
            transform.f = f;
        }
        
        if (!packed()) {
            for(Box child = getChild(0); child != null; child = child.nextSibling())
                child.place(0, 0, width, height, true);
            return;
        }
        float slack = test(XAXIS)?width:height, oldslack = 0, flex = 0, newflex = 0;
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (!child.test(VISIBLE)) continue;
            if (test(XAXIS)) {
                child._width  = child.contentWidth();
                child._height = height;
                slack -= child._width;
            } else {
                child._height  = child.contentHeight();
                child._width = width;
                slack -= child._height;
            }
            flex += child.flex;
        }
        while(slack > 0 && flex > 0 && oldslack!=slack) {
            oldslack = slack;
            slack = test(XAXIS) ? width : height;
            newflex = 0;
            for(Box child = getChild(0); child != null; child = child.nextSibling()) {
                if (!child.test(VISIBLE)) continue;
                if (test(XAXIS)) {
                    float oldwidth = child._width;
                    if (child.test(HSHRINK)) child._width = min(child.maxWidth(), child._width+(oldslack*child.flex)/flex);
                    slack -= child._width;
                    if (child._width > oldwidth) newflex += child.flex;
                } else {
                    float oldheight = child._height;
                    if (child.test(VSHRINK)) child._height = min(child.maxHeight(), child._height+(oldslack*child.flex)/flex);
                    slack -= child._height;
                    if (child._height > oldheight) newflex += child.flex;
                }
            }
            flex = newflex;
        }
        float pos = slack / 2;
        for(Box child = getChild(0); child != null; child = child.nextSibling()) {
            if (!child.test(VISIBLE)) continue;
            if (test(XAXIS)) {
                child.place(pos, 0, child._width, child._height, false);
                pos += child._width;
            } else {
                child.place(0, pos, child._width, child._height, false);
                pos += child._height;
            }
        }
    }

           
   
   

    // Rendering Pipeline /////////////////////////////////////////////////////////////////////

    private static final boolean OPTIMIZE = false;

    /** Renders self and children within the specified region. All rendering operations are clipped to xIn,yIn,wIn,hIn */
    public void render(PixelBuffer buf, Affine a, Mesh clipFrom) { render(buf, a, clipFrom, Affine.identity(), 0); }
    public void render(PixelBuffer buf, Affine a, Mesh clipFrom, Affine clipa, int bg) {
        if (!test(VISIBLE)) return;
        a = a.copy().multiply(transform);
        clipa = clipa.copy().multiply(transform);

        if (mesh == null)
            if (path != null) mesh = new Mesh(path, true);
            else {
                if (((fillcolor & 0xFF000000) != 0x00000000 || parent == null) && (text==null||"".equals(text))) {
                    mesh = new Mesh().addRect(0, 0, width, height);
                }
                // if (ret == 0) Platform.Scheduler.add(this);
                // FIXME: texture
            }
        if (mesh==null) {
            for(Box b = getChild(0); b != null; b = b.nextSibling()) b.render(buf, a, clipFrom, clipa, bg);
            if (!(text==null||text.equals("")))
                font.rasterizeGlyphs(text, buf, a.copy(), clipFrom, clipa.copy(), strokecolor, 0);
            return;
        }

        if (clipFrom != null) clipFrom.subtract(mesh, clipa);
        Mesh mesh = treeSize() > 0 ? this.mesh.copy() : this.mesh;

        if ((fillcolor & 0xff000000)!=0) bg = fillcolor;
        for(Box b = getChild(0); b != null; b = b.nextSibling()) b.render(buf, a, mesh, Affine.identity(), bg);

        mesh.fill(buf, a, null, fillcolor, true);
        if ((strokecolor & 0xff000000) != 0) mesh.stroke(buf, a, strokecolor);
        if (!(text==null||text.equals("")))
            font.rasterizeGlyphs(text, buf, a.copy(), clipFrom, clipa.copy(), strokecolor, bg);
    }
    
    // Methods to implement org.ibex.js.JS //////////////////////////////////////
  
    public JS call(JS method, JS[] args) throws JSExn {
        switch (args.length) {
            case 1: {
                //#switch(JSU.toString(method))
                case "indexof":
                    Box b = (Box)args[0];
                    if (b.parent != this)
                        return (redirect == null || redirect == this) ? JSU.N(-1) : redirect.call(method, args);
                    return JSU.N(b.getIndexInParent());

                case "distanceto":
                    Box b = (Box)args[0];
                    JS ret = new JS.Obj();
                    ret.put(JSU.S("x"), JSU.N(b.localToGlobalX(0) - localToGlobalX(0)));
                    ret.put(JSU.S("y"), JSU.N(b.localToGlobalY(0) - localToGlobalY(0)));
                    return ret;

                //#end
            }
        }
        return super.call(method, args);
    }

    public JS get(JS name) throws JSExn {
        if (JSU.isInt(name))
            return redirect == null ? null : redirect == this ? getChild(JSU.toInt(name)) : redirect.get(name);

        //#switch(JSU.toString(name))
        case "surface": return parent == null ? null : parent.getAndTriggerTraps(name);
        case "indexof": return METHOD;
        case "distanceto": return METHOD;
        case "text": return JSU.S(text);
        case "path": {
            if (path != null) return JSU.S(path.toString());
            if (text == null) return null;
            if (font == null) return null;
            String ret = "";
            for(int i=0; i<text.length(); i++) ret += font.glyphs[text.charAt(i)].path;
            return JSU.S(ret);
        }
        case "fill": return JSU.S(Color.colorToString(fillcolor));
        case "strokecolor": return JSU.S(Color.colorToString(strokecolor));
        case "textcolor": return JSU.S(Color.colorToString(strokecolor));
        case "font": return font == null ? null : font.stream;
        case "fontsize": return font == null ? JSU.N(10) : JSU.N(font.pointsize);
        case "thisbox": return this;
        case "shrink": return JSU.B(test(HSHRINK) || test(VSHRINK));
        case "hshrink": return JSU.B(test(HSHRINK));
        case "vshrink": return JSU.B(test(VSHRINK));
        case "flex": return JSU.N(flex);
        case "width": getRoot().reflow(); return JSU.N(width);
        case "height": getRoot().reflow(); return JSU.N(height);
        case "minwidth": return JSU.N(minwidth);
        case "maxwidth": return JSU.N(maxwidth);
        case "minheight": return JSU.N(minheight);
        case "maxheight": return JSU.N(maxheight);
        case "clip": return JSU.B(test(CLIP));
        case "visible": return JSU.B(test(VISIBLE) && (parent == null || (parent.get(JSU.S("visible")) == JSU.T)));
        case "axis": return test(XAXIS) ? JSU.S("x") : test(YAXIS) ? JSU.S("y") : JSU.S("z");
        case "globalx": return JSU.N(localToGlobalX(0));
        case "globaly": return JSU.N(localToGlobalY(0));
        case "cursor": return test(CURSOR) ? JSU.S((String)boxToCursor.get(this)) : null;
        case "mouse":
            if (getSurface() == null) return null;
            if (getSurface()._mousex == Integer.MAX_VALUE)
                throw new JSExn("you cannot read from the box.mouse property in background thread context");
            return new Mouse();
        case "numchildren": return redirect == null ? JSU.N(0) : redirect == this ? JSU.N(treeSize()) : redirect.get(JSU.S("numchildren"));
        case "redirect": return redirect == null ? null : redirect == this ? JSU.T : redirect.get(JSU.S("redirect"));
        case "Minimized": if (parent == null && getSurface() != null) return JSU.B(getSurface().minimized);
        default: return super.get(name);
        //#end
        throw new Error("unreachable"); // unreachable
    }

    private class Mouse extends JS.Immutable implements JS.Cloneable {
        public JS get(JS key) throws JSExn {
            //#switch(JSU.toString(key))
            case "x": return JSU.N(globalToLocalX(getSurface()._mousex));
            case "y": return JSU.N(globalToLocalY(getSurface()._mousey));

            // this might not get recomputed if we change mousex/mousey...
            case "inside": return JSU.B(test(MOUSEINSIDE));
            //#end
            return super.get(key);
        }
    }

    //#repeat setWidth/setHeight minwidth/minheight maxwidth/maxheight pendingWidth/pendingHeight
    public void setWidth(int min, int max) {
        // FIXME: deal with conflicting min/max
        if (this.minwidth == min && this.maxwidth == max) return;
        this.minwidth = min;
        this.maxwidth = max;
        RECONSTRAIN();
        if (parent != null || getSurface() == null) return;
        getSurface().pendingWidth = maxwidth;

        // FIXME: the repeat doesn't work right here
        getSurface().setMinimumSize(minwidth, minheight, minwidth != maxwidth || minheight != maxheight);
    }
    //#end

    public void put(JS name, JS value) throws JSExn {
        if (JSU.isInt(name)) { put(JSU.toInt(name), value); return; }
        //#switch(JSU.toString(name))
        case "thisbox":     if (value == null) removeSelf();
        case "text":        {
            String s = value==null ?  "" : JSU.toString(value);
            text = s;
            minwidth = maxwidth = font.textwidth(text);
            minheight = maxheight = font.textheight(text);
            RECONSTRAIN();
            dirty();
        }
        case "strokecolor": value = JSU.N(Color.stringToColor(JSU.toString(value))); CHECKSET_INT(strokecolor); dirty();
        case "textcolor":   value = JSU.N(Color.stringToColor(JSU.toString(value))); CHECKSET_INT(strokecolor); dirty();
        case "shrink":      CHECKSET_FLAG(HSHRINK | VSHRINK); RECONSTRAIN();
        case "hshrink":     CHECKSET_FLAG(HSHRINK); RECONSTRAIN();
        case "vshrink":     CHECKSET_FLAG(VSHRINK); RECONSTRAIN();
        case "path":        {
            mesh = null;
            path = new Path(JSU.toString(value));
            float tx = -1 * Encode.longToFloat2(path.horizontalBounds(Affine.identity()));
            float ty = -1 * Encode.longToFloat2(path.verticalBounds(Affine.identity()));
            path.transform(Affine.translate(tx, ty), true);
            REPLACE();
            RECONSTRAIN();
            dirty();
            polygon = null;
        }
        case "clippath":        {
            clippath = new Path(JSU.toString(value));
            //float tx = -1 * Encode.longToFloat2(clippath.horizontalBounds(Affine.identity()));
            //float ty = -1 * Encode.longToFloat2(clippath.verticalBounds(Affine.identity()));
            //clippath.transform(Affine.translate(tx, ty), true);
            REPLACE();
            RECONSTRAIN();
            dirty();
            polygon = null;
        }
        case "x":           transform.e = JSU.toInt(value);
        case "y":           transform.f = JSU.toInt(value);
        case "transform":   {
            transform = Affine.parse(JSU.toString(value));
            if (getSurface() != null)  // FIXME
                getSurface().dirty(0, 0, getSurface().root.contentwidth, getSurface().root.contentheight);
            REPLACE();
            RECONSTRAIN();
            dirty();
            polygon = null;
        }
        case "width":       setWidth(JSU.toInt(value), JSU.toInt(value));
        case "height":      setHeight(JSU.toInt(value), JSU.toInt(value));
        case "flex":        flex = JSU.toFloat(value);
        case "maxwidth":    setWidth(minwidth, JSU.toInt(value));
        case "minwidth":    setWidth(JSU.toInt(value), maxwidth);
        case "maxheight":   setHeight(minheight, JSU.toInt(value));
        case "minheight":   setHeight(JSU.toInt(value), maxheight);
        case "visible":     CHECKSET_FLAG(VISIBLE); RECONSTRAIN(); dirty();
        case "cursor":      setCursor(JSU.toString(value));
        case "fill":        setFill(value);
        case "clip":        CHECKSET_FLAG(CLIP); if (parent == null) dirty(); else parent.dirty();
        case "mouse":
            int mousex = JSU.toInt(((JS)value).get(JSU.S("x")));
            int mousey = JSU.toInt(((JS)value).get(JSU.S("y")));
            getSurface()._mousex = (int)localToGlobalX(mousex);
            getSurface()._mousey = (int)localToGlobalY(mousey);
        case "axis":        {
            String s = JSU.toString(value); 
            if (s.equals("x")) { clear(YAXIS); set(XAXIS); }
            else if (s.equals("y")) { clear(XAXIS); set(YAXIS); }
            else if (s.equals("z")) { clear(XAXIS); clear(YAXIS); }
            else JSU.warn("attempted to set axis property to invalid value: " + s);
        }
        case "Minimized": if (parent == null && getSurface() != null) getSurface().minimized = JSU.toBoolean(value);  // FEATURE
        case "Maximized": if (parent == null && getSurface() != null) getSurface().maximized = JSU.toBoolean(value);  // FEATURE
        case "Close":     if (parent == null && getSurface() != null) getSurface().dispose(true);
        case "redirect":
            for(Box cur = (Box)value; cur != null || cur == redirect; cur = cur.parent)
                if (cur == redirect) { redirect = (Box)value; return; }
            JSU.error("redirect can only be set to a descendant of its current value");
        case "fontsize": font = Font.getFont(font == null ? null : font.stream, JSU.toInt(value)); RECONSTRAIN(); dirty();
        case "font":
            if(!(value instanceof Fountain)) throw new JSExn("You can only put streams to the font property");
            //FIXME: if (font == value) return;  // FIXME: unclone()
            font = value == null ? null : Font.getFont((Fountain)value, font == null ? 10 : font.pointsize);
            RECONSTRAIN();
            dirty();
        case "titlebar": if (getSurface()!=null) getSurface().setTitleBarText(JSU.toString(value)); super.put(name,value);
        // FIXME: icon

        case "Press1":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Press2":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Press3":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release1":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release2":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Release3":      if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click1":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click2":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Click3":        if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick1":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick2":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "DoubleClick3":  if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "KeyPressed":    if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "KeyReleased":   if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "Move":          if (!test(STOP_UPWARD_PROPAGATION) && parent != null) parent.putAndTriggerTraps(name, value);
        case "HScroll":       if (!test(STOP_UPWARD_PROPAGATION) && parent != null)
            parent.putAndTriggerTraps(name, JSU.N(JSU.toFloat(value) * ((float)parent.fontSize()) / ((float)fontSize())));
        case "VScroll":       if (!test(STOP_UPWARD_PROPAGATION) && parent != null)
            parent.putAndTriggerTraps(name, JSU.N(JSU.toFloat(value) * ((float)parent.fontSize()) / ((float)fontSize())));

        case "_Move":         propagateDownward(name, value, false);
        case "_Press1":       propagateDownward(name, value, false);
        case "_Press2":       propagateDownward(name, value, false);
        case "_Press3":       propagateDownward(name, value, false);
        case "_Release1":     propagateDownward(name, value, false);
        case "_Release2":     propagateDownward(name, value, false);
        case "_Release3":     propagateDownward(name, value, false);
        case "_Click1":       propagateDownward(name, value, false);
        case "_Click2":       propagateDownward(name, value, false);
        case "_Click3":       propagateDownward(name, value, false);
        case "_DoubleClick1": propagateDownward(name, value, false);
        case "_DoubleClick2": propagateDownward(name, value, false);
        case "_DoubleClick3": propagateDownward(name, value, false);
        case "_KeyPressed":   propagateDownward(name, value, false);
        case "_KeyReleased":  propagateDownward(name, value, false);
        case "_HScroll":      propagateDownward(name, value, false);
        case "_VScroll":      propagateDownward(name, value, false);

        case "SizeChange":    return;
        case "ChildChange":   return;
        case "Enter":         return;
        case "Leave":         return;

        default:              super.put(name, value);
        //#end
    }

    private void setCursor(String value) throws JSExn {
        if (value == null) { clear(CURSOR); boxToCursor.remove(this); return; }
        if (value.equals(boxToCursor.get(this))) return;
        set(CURSOR);
        boxToCursor.put(this, value);
        Surface surface = getSurface();
        if (surface == null) return;
        String tempcursor = surface.cursor;
        propagateDownward(null, null, false);
        if (surface.cursor != tempcursor) surface.syncCursor();
    }

    private void setFill(JS value) throws JSExn {
        if (value == null) {
            if (texture == null && fillcolor == 0) return;
            texture = null;
            fillcolor = 0;
        } else if (JSU.isString(value)) {
            int newfillcolor = Color.stringToColor(JSU.toString(value));
            if (newfillcolor == fillcolor) return;
            fillcolor = newfillcolor;
            texture = null;
        } else {
            Picture newtex = Picture.load((JS)value, this);
            if (texture == newtex) return;
            texture = newtex;
            fillcolor = 0;
            if (texture != null && texture.isLoaded) run(null);
        }
        dirty();
    }

    /**
     *  Handles events which propagate down the box tree.  If obscured
     *  is set, then we merely check for Enter/Leave.
     */
    private void propagateDownward(JS name_, JS value, boolean obscured) throws JSExn {

        String name = JSU.toString(name_);
        if (getSurface() == null) return;
        int x = (int)globalToLocalX(getSurface()._mousex);
        int y = (int)globalToLocalY(getSurface()._mousey);
        boolean wasinside = test(MOUSEINSIDE);
        boolean isinside = test(VISIBLE) && inside(x, y) && !obscured;
        if (!wasinside && isinside) {
            set(MOUSEINSIDE);
            putAndTriggerTrapsAndCatchExceptions(JSU.S("Enter"), JSU.T);
        }
        if (isinside && test(CURSOR)) getSurface().cursor = (String)boxToCursor.get(this);
        if (wasinside && !isinside) {
            clear(MOUSEINSIDE);
            putAndTriggerTrapsAndCatchExceptions(JSU.S("Leave"), JSU.T);
        }

        boolean found = false;
        if (wasinside || isinside)
            for(Box child = getChild(treeSize() - 1); child != null; child = child.prevSibling()) {
                boolean save_stop = child.test(STOP_UPWARD_PROPAGATION);
                JS value2 = value;
                if (name.equals("_HScroll") || name.equals("_VScroll"))
                    value2 = JSU.N(JSU.toFloat(value) * ((float)child.fontSize()) / (float)fontSize());
                if (obscured /*|| !child.inside(x - child.x, y - child.y) FIXME FIXME */) {
                    child.propagateDownward(name_, value2, true);
                } else try {
                    found = true;
                    child.clear(STOP_UPWARD_PROPAGATION);
                    if (name != null) child.putAndTriggerTrapsAndCatchExceptions(name_, value2);
                    else child.propagateDownward(name_, value2, obscured);
                } finally {
                    if (save_stop) child.set(STOP_UPWARD_PROPAGATION); else child.clear(STOP_UPWARD_PROPAGATION);
                }
                /* FIXME FIXME
                if (child.inside(x - child.x, y - child.y))
                    if (name != null && name.equals("_Move")) obscured = true;
                    else break;
 */
            }

        if (!obscured && !found)
            if ("_Move".equals(name) || name.startsWith("_Release") || wasinside)
                if (name != null)
                    putAndTriggerTrapsAndCatchExceptions(JSU.S(name.substring(1)), value);
    }

    /** figures out what box in this subtree of the Box owns the pixel at x,y relitave to the Surface */
    public static Box whoIs(Box cur, int x, int y) {

        if (cur.parent != null) throw new Error("whoIs may only be invoked on the root box of a surface");
        int globalx = 0;
        int globaly = 0;

        // WARNING: this method is called from the event-queueing thread -- it may run concurrently with
        // ANY part of Ibex, and is UNSYNCHRONIZED for performance reasons.  BE CAREFUL HERE.

        if (!cur.test(VISIBLE)) return null;
        if (!cur.inside(x - globalx, y - globaly)) return cur.parent == null ? cur : null;
        OUTER: while(true) {
            for(int i=cur.treeSize() - 1; i>=0; i--) {
                Box child = cur.getChild(i);
                if (child == null) continue;        // since this method is unsynchronized, we have to double-check
                globalx += child.transform.e;
                globaly += child.transform.f;
                if (child.test(VISIBLE) && child.inside(x - globalx, y - globaly)) { cur = child; continue OUTER; }
                globalx -= child.transform.e;
                globaly -= child.transform.f;
            }
            break;
        }
        return cur;
    }


    // Trivial Helper Methods (should be inlined) /////////////////////////////////////////

    public final int fontSize() { return font == null ? DEFAULT_FONT.pointsize : font.pointsize; }
    public Enumeration jskeys() { throw new Error("you cannot apply for..in to a " + this.getClass().getName()); }
    public Box getRoot() { return parent == null ? this : parent.getRoot(); }
    public Surface getSurface() { return Surface.fromBox(getRoot()); }
    private final boolean packed() { return test(YAXIS) || test(XAXIS); }
    Box nextPackedSibling() { Box b = nextSibling(); return b == null || (b.packed() && b.test(VISIBLE))?b:b.nextPackedSibling(); }
    Box firstPackedChild() { Box b = getChild(0); return b == null || (b.packed() && b.test(VISIBLE))?b:b.nextPackedSibling(); }
    public float globalToLocalX(float x) { return parent == null ? x : parent.globalToLocalX(x - transform.e); }
    public float globalToLocalY(float y) { return parent == null ? y : parent.globalToLocalY(y - transform.f); }
    public float localToGlobalX(float x) { return parent == null ? x : parent.globalToLocalX(x + transform.e); }
    public float localToGlobalY(float y) { return parent == null ? y : parent.globalToLocalY(y + transform.f); }

    static short min(short a, short b) { if (a<b) return a; else return b; }
    static int min(int a, int b) { if (a<b) return a; else return b; }
    static float min(float a, float b) { if (a<b) return a; else return b; }

    static short max(short a, short b) { if (a>b) return a; else return b; }
    static int max(int a, int b) { if (a>b) return a; else return b; }
    static float max(float a, float b) { if (a>b) return a; else return b; }

    static int min(int a, int b, int c) { if (a<=b && a<=c) return a; else if (b<=c && b<=a) return b; else return c; }
    static int max(int a, int b, int c) { if (a>=b && a>=c) return a; else if (b>=c && b>=a) return b; else return c; }
    static int bound(int a, int b, int c) { if (c < b) return c; if (a > b) return a; return b; }
    final boolean inside(int x, int y) { return test(VISIBLE) && x >= 0 && y >= 0 && x < width && y < height; }

    void set(int mask) { flags |= mask; }
    void set(int mask, boolean setclear) { if (setclear) set(mask); else clear(mask); }
    void clear(int mask) { flags &= ~mask; }
    public boolean test(int mask) { return ((flags & mask) == mask); }
    

    // Tree Handling //////////////////////////////////////////////////////////////////////

    public final int getIndexInParent() { return parent == null ? 0 : parent.indexNode(this); }
    public final Box nextSibling() { return parent == null ? null : parent.getChild(parent.indexNode(this) + 1); }
    public final Box prevSibling() { return parent == null ? null : parent.getChild(parent.indexNode(this) - 1); }
    public final Box getChild(int i) {
        if (i < 0) return null;
        if (i >= treeSize()) return null;
        return (Box)getNode(i);
    }

    // Tree Manipulation /////////////////////////////////////////////////////////////////////

    public void removeSelf() {
        if (parent != null) { parent.removeChild(parent.indexNode(this)); return; }
        Surface surface = Surface.fromBox(this); 
        if (surface != null) surface.dispose(true);
    }

    /** remove the i^th child */
    public void removeChild(int i) {
        Box b = getChild(i);
        b.RECONSTRAIN();
        b.dirty();
        b.clear(MOUSEINSIDE);
        deleteNode(i);
        b.parent = null;
        RECONSTRAIN();
        putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);
    }
    
    public void put(int i, JS value) throws JSExn {
        if (i < 0) return;
            
        if (value != null && !(value instanceof Box)) {
            if (Log.on) JSU.warn("attempt to set a numerical property on a box to a non-box");
            return;
        }

        if (redirect == null) {
            if (value == null) putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), getChild(i));
            else JSU.warn("attempt to add/remove children to/from a node with a null redirect");

        } else if (redirect != this) {
            if (value != null) putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), value);
            redirect.put(i, value);
            if (value == null) {
                Box b = (Box)redirect.get(JSU.N(i));
                if (b != null) putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);
            }

        } else if (value == null) {
            if (i < 0 || i > treeSize()) return;
            Box b = getChild(i);
            removeChild(i);
            putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);

        } else {
            Box b = (Box)value;

            // check if box being moved is currently target of a redirect
            for(Box cur = b.parent; cur != null; cur = cur.parent)
                if (cur.redirect == b) {
                    if (Log.on) JSU.warn("attempt to move a box that is the target of a redirect");
                    return;
                }

            // check for recursive ancestor violation
            for(Box cur = this; cur != null; cur = cur.parent)
                if (cur == b) {
                    if (Log.on) JSU.warn("attempt to make a node a parent of its own ancestor");
                    if (Log.on) Log.info(this, "box == " + this + "  ancestor == " + b);
                    return;
                }

            if (b.parent != null) b.parent.removeChild(b.parent.indexNode(b));
            insertNode(i, b);
            b.parent = this;
            
            // need both of these in case child was already uncalc'ed
            b.RECONSTRAIN();
            RECONSTRAIN();
            
            b.dirty(); 
            putAndTriggerTrapsAndCatchExceptions(JSU.S("ChildChange"), b);
        }
    }
    
    public void putAndTriggerTrapsAndCatchExceptions(JS name, JS val) {
        try {
            putAndTriggerTraps(name, val);
        } catch (JSExn e) {
            JSU.log("caught js exception while putting to trap \""+ JSU.str(name)+"\"");
            JSU.log(e);
        } catch (Exception e) {
            JSU.log("caught exception while putting to trap \""+ JSU.str(name)+"\"");
            JSU.log(e);
        }
    }
    
    // BalancedTree functions
    private void insertNode(int p, Box b) { if(bt == null) bt = new BalancedTree(); bt.insertNode(p,b); }
    private int treeSize() { return bt == null ? 0 : bt.treeSize(); }
    private int indexNode(Box b) { return bt == null ? -1 : bt.indexNode(b); }
    private void deleteNode(int p) { bt.deleteNode(p); }
    private Box getNode(int p) { return (Box)bt.getNode(p); }

    // FIXME memory leak
    final static JS.Method METHOD = new JS.Method();
    final static Basket.Map boxToCursor = new Basket.Hash(500, 3);
    final static JS SIZECHANGE = JSU.S("SizeChange");
    final static Font DEFAULT_FONT = Font.getFont(Main.vera, 10);


    // Flags //////////////////////////////////////////////////////////////////////

    public static final int MOUSEINSIDE             = 0x00000001;
    public static final int XAXIS                   = 0x00000002;
    public static final int YAXIS                   = 0x00000004;
    public static final int VISIBLE                 = 0x00000008;
    public static final int VSHRINK                 = 0x00000010;
    public static final int HSHRINK                 = 0x00000020;
    public static final int RECONSTRAIN             = 0x00000040;
    public static final int REPLACE                 = 0x00000080;

    // if true, this box has cursor in the cursor hash; FEATURE: GC issues?
    public static final int CURSOR                  = 0x00010000;
    public static final int CLIP                    = 0x00020000;
    public static final int STOP_UPWARD_PROPAGATION = 0x00040000;
    public static final int MOVED                   = 0x00080000;


}
